/* $Id: approximate.c 140 2004-11-20 21:51:31Z ian $
 * Copyright (c) 2002 Ian Glover
 *
 * This file is part of wmfuzzy and is distributed under the GNU
 * GPL. See file COPYING for details.
 *
 * Code for taking a Unix tm structure and calculating the approximate
 * time from it.
 */
#include "approximate.h"
#include <assert.h>
#include <time.h>

static int is_leap_year(int year);

#define MINUTES_PER_DAY (60 * 24)
#define MINUTES_PER_WEEK (MINUTES_PER_DAY * 7)

static int is_leap_year(int year)
{
  return (0 == (year % 4)) && (0 != (year % 100));
}

struct ApproximateTime approximate(const struct tm* tm,
                                   const enum Granularity gr)
{
  struct ApproximateTime time;
  const double grf = (double)gr;
  double days_in_month;

  if (gr == DAY)
  {
    time.weekday = tm->tm_wday + 1;
  }
  else if (gr == WEEKTHIRD)
  {
    time.weekday
      = 1 + (int)(((double)(tm->tm_wday * MINUTES_PER_DAY
                            + tm->tm_hour * 60
                            + tm->tm_min) / (double)MINUTES_PER_WEEK) * 3.0);
  }

  if (gr == MONTHTHIRD)
  {
    if (tm->tm_mon == 3     /* April */
        || tm->tm_mon == 5  /* June */
        || tm->tm_mon == 8  /* September */
        || tm->tm_mon == 10 /* November */)
    {
      days_in_month = 30.0;
    }
    else if (tm->tm_mon == 1) /* February */
    {
      if (is_leap_year(tm->tm_year)) 
      {
        days_in_month = 29.0;
      }
      else
      {
        days_in_month = 28.0;
      }
    }
    else
    {
      days_in_month = 31.0;
    }
    
    time.monththird = 1 + (int)(((double)(tm->tm_mday) / days_in_month) * 3.0);
  }
  else { time.monththird = 0; }

  if (gr >= SIXHOURS)
  {
    time.year = tm->tm_year + 1900;
    time.month = tm->tm_mon + 1;
    time.hour = tm->tm_hour;
    time.minutes = 0;
    time.join = TO;
    return (time);
  }
 
  time.hour = tm->tm_hour;

  /* Round to the nearest minute rather than just down. Does this
     actually work on a compiler that doesn't round integer division
     down? Are there any such and does it really matter? */
  time.minutes = (int)(((double)(tm->tm_min) + 0.5 * grf) / grf) * gr;

  if (time.minutes > 30)
  {
    time.hour += 1;
    time.join = TO;
  }
  else
  {
    time.join = PAST;
  }

  if (time.minutes > 30) {
    time.minutes = 60 - time.minutes;
    assert(time.minutes >= 0);
  }
  time.hour    %= 12;

  if (time.hour == 0) {
    time.hour = 12;
  }

  return (time);
}
