/* $Id: converttotext.c 144 2004-11-22 20:14:08Z ian $
 * Copyright (c) 2002-2003 Ian Glover
 *
 * This file is part of wmfuzzy and is distributed under the GNU
 * GPL. See file COPYING for details.
 */
#include "converttotext.h"
#include "approximate.h"
#include MESSAGE_HEADER
#include "util.h"
#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#define TEXT_LENGTH 40

static int calculate_century(int year);
static void text_from_table(int index,
                            const char* table[], size_t table_length,
                            char* text, size_t len);
static void find_suffix(int ordinal, char* text, size_t len);

static void year_to_text(const int year, const enum Granularity gran,
                         char* text, const size_t len);
static void season_to_text(const int month, char* text, const size_t len);
static void month_to_text(const int month, char* text, const size_t len);
static void monththird_to_text(const int monththird,
                               char* text, const size_t len);
static void weekthird_to_text(const int third, char* text, const size_t len);
static void weekday_to_text(const int weekday, char* text, const size_t len);
static void hour_to_day_quarter(const int hour, char* text, const size_t len);
static void hour_to_day_half(const int hour, char* text, const size_t len);
static void number_to_text(const int number, char* text, const size_t len);
static void minute_to_text(const int minute, char* text, const size_t len);

static void write_hour(const int hour, char* text, size_t len);
static void write_minutes(const struct ApproximateTime* approximation,
                          char* text, size_t len);


void year_to_text(const int year, const enum Granularity gran,
                  char* text, const size_t len)
{
  if(gran == YEAR)
  { 
    (void)snprintf(text, len, "%i", year);
  }
  else if(gran == CENTURY)
  {
    char suffix[3];
    memset(suffix, 0, ARRAY_SIZE(suffix));
    find_suffix(calculate_century(year), suffix, ARRAY_SIZE(suffix));
    
    assert(len > 12);
    (void)snprintf(text, len, "%i%s %s", calculate_century(year), suffix,
		   text_century);
  }
}

/* Convert a month into a season name. */
void season_to_text(const int month, char* text, const size_t len)
{
  const int season = (month % 12) / 3;

  text_from_table(season, season_names, ARRAY_SIZE(season_names), text, len);
}

/* Convert a month number into the month's name. */
void month_to_text(const int month, char* text, const size_t len)
{
  text_from_table(month - 1, month_names, ARRAY_SIZE(month_names), text, len);
}

void monththird_to_text(const int third, char* text, const size_t len)
{
  text_from_table(third - 1, third_names, ARRAY_SIZE(third_names),
                  text, len);
}

/* Convert a weekday number to a name. */
void weekday_to_text(const int weekday, char* text, const size_t len)
{
  text_from_table(weekday - 1, weekday_names, ARRAY_SIZE(weekday_names),
                  text, len);
}

void weekthird_to_text(const int third, char* text, const size_t len)
{
  text_from_table(third - 1, week_thirds, ARRAY_SIZE(week_thirds), text, len);
}

void hour_to_day_quarter(const int hour, char* text, const size_t len)
{
  const int period = ((int)(hour / 6));
  
  assert(hour >= 0);
  assert(hour <= 23);

  text_from_table(period,
                  quarter_day_periods, ARRAY_SIZE(quarter_day_periods),
                  text, len);
}

void hour_to_day_half(const int hour, char* text, const size_t len)
{
  assert(hour >= 0);
  assert(hour <= 23);

  text_from_table( (hour >= 6 && hour < 18) ? 0 : 1,
                   half_day_periods, ARRAY_SIZE(half_day_periods),
                   text, len);
}
  
void number_to_text(const int number, char* text, const size_t len)
{
  
  assert((size_t)number < ARRAY_SIZE(number_names));
  assert(number >= 0);

  text_from_table(number, number_names, ARRAY_SIZE(number_names), text, len);
}  

/* Convert a minute rounded to the nearest multiple of five to name
 * for that past the hour. */
void minute_to_text(const int minute, char* text, const size_t len)
{
  assert(minute % 5 == 0);
  assert(minute <= 30);
  assert(minute > 0);

  text_from_table(minute / 5 - 1, minute_names, ARRAY_SIZE(minute_names),
                  text, len);
}

void write_hour(const int hour, char* text, size_t len)
{
  char buffer[30];
  memset(buffer, 0, ARRAY_SIZE(buffer));
  
  number_to_text(hour, buffer, ARRAY_SIZE(buffer));
  snprintf(text, len, "%s %s", buffer, text_oclock);
}

void write_minutes(const struct ApproximateTime* approximation,
                   char* text, size_t len)
{
  char minute_text[15];
  char hour_text[15];
  const char* join;
  memset(minute_text, 0, ARRAY_SIZE(minute_text));
  memset(hour_text, 0, ARRAY_SIZE(hour_text));    
  
  if (approximation->join == TO)
  {
    join = text_join_before;
  }
  else
  {
    assert(approximation->join == PAST);
    join = text_join_after;
  }
  minute_to_text(approximation->minutes, minute_text,
                 ARRAY_SIZE(minute_text));
  number_to_text(approximation->hour, hour_text, ARRAY_SIZE(hour_text));

  if (approximation->minutes > 30)
  {    
    fmt_after_half_hour(text, len, hour_text, minute_text, join);
  }
  else
  {
    fmt_before_half_hour(text, len, hour_text, minute_text, join);    
  }
}

/* Given an index and a table of strings of a given size, return the
   string pointed to by that index or null if there is no such. */
static void text_from_table(int index,
                            const char* table[],
                            size_t table_length,
                            char* buffer,
                            size_t len)
{
  assert(index >= 0 );
  assert(((size_t)index) < table_length);

  if ((index >= 0) && (((size_t)index) < table_length))
  {
    strncpy(buffer, table[index], len);
  }
}

/* Return the century that a year is in. Probably won't work for BC. */ 
static int calculate_century(int year)
{
  return (int)(year / 100) + 1;
}

/* Take the index of a ordinal and return the suffix for it, so for
 * example 1 -> st . */
static void find_suffix(int ordinal, char* text, size_t len)
{
  int suffix = (ordinal % 10) - 1;
  if ( suffix > 3 ) { suffix = 3; }
  
  text_from_table(suffix, suffix_table, ARRAY_SIZE(suffix_table), text, len);
}

void write_time(const struct tm* tm, const enum Granularity gran,
                char* text, const int len)
{
  const struct ApproximateTime approximation = approximate(tm, gran);

  memset(text, 0, len);  

  if (approximation.minutes == 0)
  {
    if(gran <= HOUR)
    {
      write_hour(approximation.hour, text, len);
    }
    else if(gran == MONTHTHIRD)
    {
      char third_buffer[TEXT_LENGTH];
      char month_buffer[TEXT_LENGTH];
      memset(third_buffer, 0, TEXT_LENGTH);
      memset(month_buffer, 0, TEXT_LENGTH);

      
      monththird_to_text(approximation.monththird, third_buffer,
                         TEXT_LENGTH);
      month_to_text(approximation.month, month_buffer,
                    TEXT_LENGTH);
      snprintf(text, len, "%s %s", third_buffer, month_buffer);
    }
    else
    {
      switch (gran)
      {
        case SIXHOURS:
          hour_to_day_quarter(approximation.hour, text, len);
          break;
        case TWELVEHOURS:
          hour_to_day_half(approximation.hour, text, len);
          break;
        case DAY:
          weekday_to_text(approximation.weekday, text, len);
          break;
        case WEEKTHIRD:
          weekthird_to_text(approximation.weekday, text, len);
          break;          
        case MONTH:
          month_to_text(approximation.month, text, len);
          break;
        case SEASON:
          season_to_text(approximation.month, text, len);
          break;
        case YEAR:
          year_to_text(approximation.year, gran, text, len);
          break;
        case CENTURY:
          year_to_text(approximation.year, gran, text, len);
          break;
        default:
          assert(0);
          break;
      }
    }
  }
  else
  {
    write_minutes(&approximation, text, len);
  }
}
