/* $Id: wmfuzzy.c 141 2004-11-21 08:42:56Z ian $
 * Copyright (c) 2002-2004 Ian Glover
 *
 * This file is part of wmfuzzy and is distributed under the GNU
 * GPL. See file COPYING for details.
 */
#include "wmfuzzy.h"
#include "approximate.h"
#include "converttotext.h"
#include "logging.h"
#include "util.h"
#include "version.h"
#include "xdisplay.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define NAME "wmFuzzy"

#define TEXT_LENGTH 30

enum Granularity granularity = FIVE;

static void set_granularity(char* txt);

static void show_help(void);
static void show_version(void);

static void do_update(void);
static void do_click(void);

static void dump_time(FILE* f);

/* Function definition required for all the callback functions. */
typedef void (*Callback)(void);

static /*@null@*/ Callback update_fn;
static /*@null@*/ Callback clicked_fn;

static void lib_initialise(void);

/* This registers callback for when the windowed is clicked. */
static void set_clicked_callback(Callback fn);

static void set_update_callback(Callback fn);

/* This function should be supplied by the dockApp application code to
   perform initialisation, registering callbacks etc. */
static void app_setup(int argc, char* argv[]);

void set_granularity(char* txt)
{
  long value;
  granularity = FIVE;

  if (txt == 0) { assert(0); return; }
  value = strtol(txt, 0, 10);

  down_case(txt);
  if (strcmp(txt, "century") == 0) { granularity = CENTURY; }
  else if (strcmp(txt, "year") == 0) { granularity = YEAR; }
  else if (strcmp(txt, "season") == 0) { granularity = SEASON; }
  else if (strcmp(txt, "month") == 0) { granularity = MONTH; }
  else if (strcmp(txt, "day") == 0) { granularity = DAY; }
  else
  {
    if (errno == ERANGE || errno == EINVAL)
    {
      warning("Error process -g argument value : %s\n", txt);
    }
    else if (value < 7) { granularity = FIVE; }
    else if (value < 12) { granularity = TEN; }
    else if (value < 17) { granularity = FIFTEEN; }
    else if (value < 22) { granularity = TWENTY; }
    else if (value < 45) { granularity = HALFHOUR; }
    else if (value < 210) { granularity = HOUR; }
    else if (value < 540) { granularity = SIXHOURS; }
    else if (value < 1080) { granularity = TWELVEHOURS; }
    else if (value < 2400) { granularity = DAY; }
    else if (value < 8880) { granularity = WEEKTHIRD; }
    else if (value < 28800) { granularity = MONTHTHIRD; }
    else if (value < 86400) { granularity = MONTH; }
    else if (value < 327600) { granularity = SEASON; }
    else if (value < 26542800) { granularity = YEAR; }
    else { granularity = CENTURY; }
  }
}

void show_help(void) {
  show_version();
  
  printf(
    "\n"
    "-b<colour> :  Set the background colour. See -f.\n"
    "-d         :  Print the current approximation and exit.\n"
    "-f<colour> :  Set the foreground colour, either as a name or in the\n"
    "              form #RRGGBB where R,G and B are hexidecimal characters.\n"

    "-g<n>      :  Set the granularity, n should be one of 5, 10, 15, 20,\n"
    "              30, 60, 360, 720, 1440, 3360, 14400, 43200, 129600, 525600\n"
    "              or 52560000. Other values will be approximated to the\n"
    "              nearest of these. Alternatively the words \"century\",\n"
    "              \"year\", \"month\", or \"day\" can be used as shortcuts\n"
    "              for those settings. The default is 5 minutes.\n"
    "-h         :  Print this message and exit.\n"
    "-v         :  Print version information and exit.\n");
}

void show_version(void)
{
  printf("%s, version %s by Ian Glover (ian@manicai.net)\n", NAME, VERSION);
}

void do_update(void)
{
  time_t seconds;
  struct tm time;
  static char text[TEXT_LENGTH];
  
  seconds = get_time();
  localtime_r(&seconds, &time);
  write_time(&time, granularity, text, TEXT_LENGTH);
  display_set_text(text, TEXT_LENGTH);
  display_set_update_delay(30);
}


void do_click(void)
{
  static char line_buffer[TEXT_LENGTH];
  time_t seconds;
  struct tm time;

  memset(line_buffer, 0, TEXT_LENGTH);

  seconds = get_time();
  localtime_r(&seconds, &time);

  strftime(line_buffer, TEXT_LENGTH, "%H:%M %d_%b %Y", &time);

  display_set_text(line_buffer, TEXT_LENGTH);
  display_set_update_delay(5);
}

void dump_time(FILE* f)
{
  time_t seconds = get_time();
  struct tm time;
  static char text[TEXT_LENGTH];

  memset(text, 0, TEXT_LENGTH);
  localtime_r(&seconds, &time);
  write_time(&time, granularity, text, TEXT_LENGTH);
  fprintf(f, "%s\n", text);
}

void app_setup(int argc, char** argv)
{
  int c;
  int dump = 0;
  
  while( -1 != (c = getopt(argc, argv, "b:df:g:hv"))) {
    assert(c <= (int)CHAR_MAX);
    assert(c >= (int)CHAR_MIN);
      
    switch((char)c) {
      case '?': exit(EXIT_FAILURE);
      case 'b': display_set_background(optarg); break;
      case 'd': dump = 1; break;
      case 'f': display_set_foreground(optarg); break;
      case 'h':
      {
        show_help();
        exit(EXIT_SUCCESS);
      }
      case 'g': set_granularity(optarg); break;
      case 'v':
      {
        show_version();
        exit(EXIT_SUCCESS);        
      }
      default:
      {
        fatal("Unhandled command line argument.");
        break;
      }
    }
  }

  // Print the time to stdout and exit.
  if ( dump )
  {
    dump_time(stdout);
    exit(EXIT_SUCCESS);
  }
  
  set_update_callback(&do_update);
  set_clicked_callback(&do_click);
  
  display_set_app_name("wmFuzzy");
  
  display_set_update_delay(30); /* seconds */
}

void lib_initialise(void)
{
  update_fn = NULL;
  clicked_fn = NULL;
}

void call_update_fn(void)
{
  if (NULL != update_fn)
  {
    (*update_fn)();
  }
}

void call_clicked_fn(void)
{
  if (NULL != clicked_fn)
  {
    (*clicked_fn)();
  }
}

void set_update_callback(Callback fn)
{
  update_fn = fn;
}

void set_clicked_callback(Callback fn)
{
  clicked_fn = fn;
} 

int main(int argc, char* argv[])
{
  lib_initialise();
  display_initialise(argc, argv);

  display_set_app_name(argv[0]);
  app_setup(argc, argv);    
  
  display_main_loop();
  display_free_resources();
  
  return 0;
}
