/* $Id: xdisplay.c 141 2004-11-21 08:42:56Z ian $
 * Copyright (c) 2003-2004 Ian Glover
 *
 * This file is part of wmfuzzy and is distributed under the GNU
 * GPL. See file COPYING for details.
 */

#include "xdisplay.h"
#include "wmfuzzy.h"
#include "wmmask.xbm"
#include "wmicon.xpm"
#include "logging.h"
#include "util.h"

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h> /* for alarm and usleep. */
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#define MAX_NUM_LINES 4
#define MAX_LINE_LENGTH 9

char* x_msg_lines[MAX_NUM_LINES];

static void alarm_handler(int sig_num);

static void x_create_window_gc(void);
static int  x_flush_expose(Window w);
static void x_redraw_window(XEvent* event);
static void x_create_window(void);
static void x_display_window(void);
static void x_print_message(void);
static int  x_get_colour(const char* name, XColor* colour);
static void x_handle_event(XEvent* event);
static void x_set_message(char* lines[], int nLines);

void do_click(void);

typedef struct {
	Pixmap		pixmap;
	Pixmap		mask;
	XpmAttributes	attributes;
} XpmIcon;

static Display* display;
static Window iconwin;
static Window window;
static GC xgc;
static XpmIcon wmbg;

static unsigned int update_wait = 60;

struct Settings {
  XColor colour_fg;
  XColor colour_bg;
};

static struct Settings config;

static sig_atomic_t alarm_call = 0;
static int complete_refresh = 0;

/*
 * Handle alarm calls
 */
void alarm_handler(int /*@unused@*/ sig_num)
{
  ++alarm_call;
}

/*
 * Create a graphics context for the display.
 */
void x_create_window_gc(void)
{
  XColor colour;
  XGCValues values;
  values.graphics_exposures = 0;

  (void)x_get_colour("white", &colour);
  values.foreground = colour.pixel;

  (void)x_get_colour("black", &colour);
  values.background = colour.pixel;
  
  xgc = XCreateGC(display,
                  RootWindow(display, DefaultScreen(display)),
                  GCForeground | GCBackground | GCGraphicsExposures,
                  &values);
    
  if (0 == xgc)
  {
    fatal("Cannot create X graphics context.\n");
  }	
}

/*
 * Do the dirty work of printing the text.
 */
void x_print_message(void)
{
  int i;
  int vtab = 15;
  static const int half_char_height = 7;
  static const int half_char_width = 3;
  
  for (i = 0; i < MAX_NUM_LINES; ++i)
  {
    if (0 == x_msg_lines[i]) { vtab += half_char_height; }
  }
  
  for (i = 0; i < MAX_NUM_LINES; i++)
  {    
    if (0 != x_msg_lines[i])
    {
      int htab;
      size_t len = strlen(x_msg_lines[i]);
      
      if (len > 8) { len = 8; }
      htab = 32 - half_char_width * len;
      
      (void)XDrawString(display, wmbg.pixmap, xgc, htab, vtab,
                        x_msg_lines[i], len);
      vtab += 2 * half_char_height;
    }
  }
}

int x_flush_expose(Window w)
{
  XEvent dummy_event;
  int i = 0;

  while (XCheckTypedWindowEvent(display, w, Expose, &dummy_event))
  {
    ++i;
  }
  return i;
}

void x_redraw_window(/*@null@*/ XEvent* event)
{

  (void)XSetForeground(display, xgc, config.colour_bg.pixel);
  /* Having the dimensions hard code is a bit nasty. */
  (void)XFillRectangle(display, wmbg.pixmap, xgc, 5, 5, 54, 54);
  
  (void)XSetForeground(display, xgc, config.colour_fg.pixel);

  x_print_message();

  /* May want to combine these clauses to just use the code currently
   * and makes things simpler in the else. But doing so makes the graphs
   * slightly more flickery. May also relate to OpenBSD problem?*/
  if (NULL != event && !complete_refresh)
  {
    /* Using x_flush_expose prior to these calls means that only the
       first (or last?)  expose event is noticed so we get some ugly
       white chunks. */
    /* x_flush_expose(window); */
    (void)XCopyArea(display, wmbg.pixmap, iconwin, xgc,
                    event->xexpose.x, event->xexpose.y,
                    event->xexpose.width, event->xexpose.height,
                    event->xexpose.x, event->xexpose.y);
    /* x_flush_expose(window); */
    (void)XCopyArea(display, wmbg.pixmap, window, xgc,
                    event->xexpose.x, event->xexpose.y,
                    event->xexpose.width, event->xexpose.height,
                    event->xexpose.x, event->xexpose.y);
  }
  else
  {
    (void)x_flush_expose(iconwin);
    (void)XCopyArea(display, wmbg.pixmap, iconwin, xgc, 0, 0,
                    wmbg.attributes.width, wmbg.attributes.height, 0, 0);
    (void)x_flush_expose(window);
    (void)XCopyArea(display, wmbg.pixmap, window, xgc, 0, 0,
                    wmbg.attributes.width, wmbg.attributes.height, 0, 0);
    
    complete_refresh = 0;
  }

}

int x_get_colour(const char* name, XColor* colour) {
  Colormap colourmap;

  colourmap = DefaultColormap(display, DefaultScreen(display)); 

  if(!XParseColor(display, colourmap, name, colour))
  {
    warning("Couldn't parse \"%s\"\n", name);
    return 0;
  }
  if(!XAllocColor(display, colourmap, colour))
  {
    warning("Couldn't alloc \"%s\"\n", name);
    return 0;
  }
	
  return 1;
}

void x_create_window(void) {
  XWMHints wmHints;
  Pixmap pixmask;
	
  /* connect to x server */
  if((display = XOpenDisplay(NULL)) == NULL)
  {
    fatal("Couldn't open display\n");
  }
	
  /* create the window */
  window = XCreateSimpleWindow(display,
                               RootWindow(display, 0),
                               0, 0, 64, 64, 1,
                               BlackPixel(display, DefaultScreen(display)),
                               WhitePixel(display, DefaultScreen(display)));
  iconwin = XCreateSimpleWindow(display, window, 0, 0, 64, 64, 1,
                                BlackPixel(display, DefaultScreen(display)),
                                WhitePixel(display, DefaultScreen(display)));

  /* put pixmask over window so that outer border becomes transparent */
  pixmask = XCreateBitmapFromData(display, window, wmmask_bits,
                                  wmmask_width, wmmask_height);
  XShapeCombineMask(display, window, ShapeBounding, 0, 0, pixmask, ShapeSet);
  XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
	
  /* make it a dockapp */
  wmHints.initial_state = WithdrawnState;
  wmHints.window_group = window;
  wmHints.icon_window = iconwin;
  wmHints.icon_x = 0;
  wmHints.icon_y = 0;
  wmHints.flags = (StateHint | WindowGroupHint | IconWindowHint
                   | IconPositionHint);
  (void)XSetWMHints(display, window, &wmHints);
	
  /* require event-notification */
  (void)XSelectInput(display, window,
                     ExposureMask | PointerMotionMask | ButtonReleaseMask
                     | ButtonPressMask | StructureNotifyMask);
  (void)XSelectInput(display, iconwin,
                     ExposureMask | PointerMotionMask | ButtonReleaseMask
                     | ButtonPressMask | StructureNotifyMask);


  x_create_window_gc();
  assert(xgc);    
  
  if (XpmCreatePixmapFromData(display, window,
                              wmicon_xpm,
                              &wmbg.pixmap,
                              &wmbg.mask,
                              &wmbg.attributes) != XpmSuccess)
  {
    fatal("Could not open background.\n");
  }  
}

void x_display_window(void)
{ 
  (void)XMapWindow(display, window);
  (void)XFlush(display);
}

void x_handle_event(XEvent* event) 
{
  if (0 == event)
  {
    return;
  }
  
  switch (event->type)
  {
    case Expose:
      x_redraw_window(event);
      break;
    case ButtonPress:
      call_clicked_fn(); 
      x_redraw_window(0);
      break;
    default:
      /* Don't know what's happened but we can redraw the window just
       * in case. */
      x_redraw_window(0);      
      break;
  }
}

void x_set_message(char* lines[], int nLines)
{
  int i;
  
  if (nLines > MAX_NUM_LINES) { nLines = MAX_NUM_LINES; }

  for (i = 0; i < nLines; ++i)
  {
    if (0 != x_msg_lines[i])
    {
      free(x_msg_lines[i]);
    }
    x_msg_lines[i] = malloc(strlen(lines[i]) + 1);
    memset(x_msg_lines[i], 0, strlen(lines[i] + 1));
    strcpy(x_msg_lines[i], lines[i]);
  }
  for ( ; i < MAX_NUM_LINES; ++i)
  {
    x_msg_lines[i] = 0;
  }
}

/*
 * Publically visible functions.
 */
void display_free_resources(void)
{
  Colormap cmap = DefaultColormap(display, DefaultScreen(display)); 
  (void)XFreeColors(display, cmap, &config.colour_fg.pixel, 1, 0);
  (void)XFreeColors(display, cmap, &config.colour_bg.pixel, 1, 0);

  (void)XCloseDisplay(display);
}

void display_initialise(int argc, char* argv[])
{
  int i;

  for (i = 0; i < MAX_NUM_LINES; ++i)
  {
    x_msg_lines[i] = 0;
  }
  
  x_create_window();
  if (0 == window)
  {
    fatal("Cannot create window.\n");
  }
  (void)XSetCommand(display, window, argv, argc);

  (void)x_get_colour("black", &config.colour_bg);
  (void)x_get_colour("green", &config.colour_fg);
}

void display_main_loop(void)
{
  XEvent event;
  /* Display the window */
  x_display_window();

  call_update_fn();
  (void)signal(SIGALRM, &alarm_handler);
  (void)alarm(update_wait);
  
  while (1)
  {
    if (alarm_call > 0)
    {
      call_update_fn();

      --alarm_call;
      (void)alarm(update_wait);
      complete_refresh = 1;
    }
    
    if (1 == complete_refresh)
    {
      x_redraw_window(0);
    }

    while (XPending(display))
    {
      XNextEvent(display, &event);
      x_handle_event(&event);
    }

    usleep(100);
  }
}

void display_set_app_name(const char* name)
{
  XClassHint classHint;
  
  static char* lName = 0;
  if (0 != lName)
  {
    free(lName);
  }
  lName = malloc(strlen(name) + 1);
  assert(lName);
  memset(lName, 0, strlen(name) + 1);
  strcpy(lName, name);

  classHint.res_name = lName;
  classHint.res_class = lName;
  XSetClassHint(display, window, &classHint);
}

void display_set_update_delay(int seconds)
{
  update_wait = seconds;
  (void)alarm(update_wait);
}

void display_set_foreground(const char* colour)
{
  XColor new_fg;
  Colormap colourmap = DefaultColormap(display, DefaultScreen(display)); 

  if (x_get_colour(colour, &new_fg))
  {
    (void)XFreeColors(display, colourmap, &config.colour_fg.pixel, 1, 0);
    config.colour_fg = new_fg;
  }

  complete_refresh = 1;
}

void display_set_background(const char* colour)
{
  XColor new_bg;
  Colormap colourmap = DefaultColormap(display, DefaultScreen(display)); 

  if (x_get_colour(colour, &new_bg))
  {
    (void)XFreeColors(display, colourmap, &config.colour_bg.pixel, 1, 0);
    config.colour_bg = new_bg;
  }

  complete_refresh = 1;
} 

void display_set_text(const char* text, size_t text_length)
{
  int line_number = 0;
  size_t line_length;

  /* Text will be copied into raw_buffer and formatted correctly. +1
     is for the null character string terminator. */
  static char raw_buffer[(MAX_LINE_LENGTH + 1) * MAX_NUM_LINES];
  /* Lines will be built out of line_buffer. */
  static char line_buffer[(MAX_LINE_LENGTH + 1) * MAX_NUM_LINES];
  static char* line_text[MAX_NUM_LINES];
  
  const char* current = raw_buffer;
  const char* last = raw_buffer;
  int i = 0;

  memset(raw_buffer, 0, ARRAY_SIZE(raw_buffer));
  memset(line_buffer, 0, ARRAY_SIZE(line_buffer));
  memset(line_text, 0, ARRAY_SIZE(line_text));
  
  /* Set up the line pointrs into line_buffer. */
  for (i = 0; i < MAX_NUM_LINES; ++i )
  {
    line_text[i] = line_buffer + i * (MAX_LINE_LENGTH + 1);
  }
  
  /* text must be formatted to end with a space. Copy it into a new
     buffer to ensure this. */
  assert(strlen(text) <= text_length);
  assert(ARRAY_SIZE(raw_buffer) > text_length + 1);
  strcpy(raw_buffer, text);  
  assert(strlen(raw_buffer) < ARRAY_SIZE(raw_buffer) - 1);
  raw_buffer[strlen(raw_buffer)] = ' ';
  assert(strlen(raw_buffer) < ARRAY_SIZE(raw_buffer));  

  /* This has to be true to allow us space for the null characters to
     terminate the lines. */
  assert(ARRAY_SIZE(line_buffer) > text_length + MAX_NUM_LINES);
  /* Clear the line buffer. */  
  assert(line_buffer == line_text[0]);

  /* Put each word on it's own line. */
  while ((NULL != (current = strchr(last, ' ')))
         && (line_number < MAX_NUM_LINES))
  {
    assert(0 != current);
    assert(0 != last);
    assert(current > last);
    assert((size_t)(current - last) < text_length);
    assert(0 != line_text[line_number]);
           
    line_length = (size_t)(current - last);
    if (line_length > MAX_LINE_LENGTH)
    {
      line_length = MAX_LINE_LENGTH;
    }

    strncpy(line_text[line_number], last, line_length);

    assert(toupper(line_text[line_number][0]) < CHAR_MAX);

    line_text[line_number][0] = toupper(line_text[line_number][0]);

    last = current + 1;
    ++line_number;
  }

  /* Substitute spaces for underscores. */
  for (i = 0; i < ARRAY_SIZE(line_buffer); ++i)
  {
    if (line_buffer[i] == '_')
    {
      line_buffer[i] = ' ';
    }
  }
      
  
  x_set_message(line_text, line_number);
}
